# FingerPullRequest
### When a simple PullRequest can lead to hidden and unexpected side effects.

`FingerPullRequest` is a tool that allows you to write custom scripts that execute based on the PRs in a repository.
You can write different kind of automation, for example:
- Automatic `CHANGELOG.md` update.
- Send weekly recaps on Slack.
- Run Jira automations based on PR status and content.
- Update the Github PR template keeping it in sync with the markdown syntax that powers this script.

## Key Features

### Error Reporting
If an error occurs, a Slack message will be sent to the requested channel.

### Pull Request
The `PullRequest` type models a GitHub Pull Request, and exposes all its properties. You can use it to perform work/logic parametric on such content.

### Markdown parser and generator
Under the hood, this tool is powered by a markdown parser. This gives you the ability to define and parse markdown structures that can be used to parse content or to write template files.
To automatize workflows, a developer often need to manually insert structured pieces of info – like `Checkbox`s or `TextInput`s – into the body of PR and devise a strategy to fetch such info.

`FingerPullRequest` allows you to parse and generate these items without any effort :tada:
example:
``` swift
import FingerPullRequest

let changelogText = MarkdownElements.TextInput("Changelog Entry:", description: "Insert here the changelog")
let updateChangelog = MarkdownElements.Checkbox("Should update the changelog")

let markdown = """
 - [x] Should update the changelog
 ### Changelog Entry
 > Insert here the changelog
 - My custom changelog text.
"""

let document = MarkdownDocument(parsing: markdown)
print(document.getValue(updateChangelog)) // Prints: true
print(document.getValue(changelogText)) // Prints: `My custom changelog text.`
```
This is very handy to parse informations inserted in the PRs, for this reason there is the
 `.markdown: MarkdownDocument` property in `PullRequest` struct.
 
Every `MarkdownElements` it's able to generate its own markdown representation, via the `.markdown: String` property:
```
 import FingerPullRequest

let changelogText = MarkdownElements.TextInput("Changelog Entry:", description: "Insert here the changelog")
print(changelogText.markdown)
// Prints
// ### Changelog Entry
// > Insert here the changelog
// - 

let updateChangelog = MarkdownElements.Checkbox("Should update the changelog")
print(updateChangelog.markdown)
// Prints
// - [ ] Should update the changelog
```
The generated markdown is the template that can be used to eventually collect any expected input from the user. 

## Install
- TODO: when this tool is moved in the monorepo

## How do you use it ?

Simply create a `FingerPullRequest` instance and call the `.run` method selecting on which PRs you want to execute your logic.
example:
```swift
let fingerPullRequest = try FingerPullRequest.live(errorReporterSlackChannel: "xxx")

let result = try fingerPullRequest.run(on: .merged(.recenlty(limit: 100))) { pr 
  print(pr.title)
}
```

## Plugins
`FingerPullRequest` ships with some built-in plugins covering the most common use cases.

### Danger
A plugin to validate a PR in Danger.

```swift
let dangerResult = try fingerPullRequest.dangerCheck(prNumber: FingerPullRequest.pullRequestBuiltInCI) { pr in
   // If the PR is invalid simply throw an error, it will make the check fail and add a comment in the PR with the error
}
```

### UpdateChangelog
This plugin updates the `CHANGELOG.md` based on all the PRs merged after the last time has ran.
To know which is the last time that this plugin has run, it checks in the git history for the last
`commitMessage` generated by the Changelog plugin.
If no such commit, the plugin runs only on the most recent merge.

```swift
let updateChangelogResult = try fingerPullRequest.updateChangelog(
  file: .init(fileURLWithPath: "../../CHANGELOG.md"),
  commitMessage: "My custom commit message"
) { pr in
  // Return a String based on the PR that will become the Changelog entry
  return "Merged PR [#\(pr.number)](https://github.com/BendingSpoons/splice-iOS-swift/pull/\(pr.number))"
}
```

### UpdatePullRequestTemplate
It's important that the `MarkdownElements` used are in sync with the ones present in the PR template.
To enforce this, we've added a plugin that generates the markdown, and commits it if it differs from the one
already present in the repo.
```swift
let changelogText = MarkdownElements.TextInput("Changelog Entry:", description: "Insert here the changelog")
let updateChangelog = MarkdownElements.Checkbox("Should update the changelog")

let template = """
# My custom PR template

\(changelogText.markdown)

\(updateChangelog.markdown)
"""

let updatePullRequestTemplate = try try fingerPullRequest.updatePullRequestTemplate(
  file: .init(fileURLWithPath: "../../.github/pull_request_template.md"),
  commitMessage: "My custom commit message"
  template: template
)
```
