// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// This is the container for all the dependencies used by `FingerPullRequest`
public struct Environment {
  /// The fileSystem dependency
  public let fileSystem: FileSystem

  /// The shell dependency
  public let shell: Shell

  /// The git dependency
  public let git: Git

  /// The network dependency
  public let network: Network

  /// The github dependency
  public let github: Github

  /// The slack dependency
  public let slack: Slack

  /// The errorReporter dependency
  public let report: ErrorReporter
}

extension Environment {
  /// Returns a live instance of the `Environment`
  /// - Parameters:
  ///   - errorReporterSlackChannel: The Slack channel where the runtime errors are notified.
  ///   - errorResponsiblePeople: The list of handles of the responsible people to tag in case of a runtime error.
  ///   - githubAuthToken: The Github access token.
  ///   - slackAccessToken: The Slack access token
  ///   - repositoryName: The repositoryName, if nil is extracted automatically from Git.
  /// - Returns: The `.live` environment.
  public static func live(
    errorReporterSlackChannel channel: String,
    errorResponsiblePeople: [String],
    githubAuthToken: String,
    slackAccessToken: String,
    repositoryName: String?
  ) -> Self {
    let shell: Shell = .bash()
    let fileSystem: FileSystem = .live()
    let network: Network = .live(fileSystem: fileSystem)
    let git: Git = .live(shell: shell)
    let slack: Slack = .live(network: network, slackAccessToken: slackAccessToken)
    let repository = repositoryName ?? (try! git.remoteURL()!.deletingPathExtension().path)
    let errorMessageComposer: ErrorMessageComposer = .live(tags: errorResponsiblePeople)
    let errorReporter: ErrorReporter = .live(
      slack: slack,
      composeErrorMessage: errorMessageComposer,
      channel: channel
    )
    let github: Github = .live(network: network, repository: repository, authToken: githubAuthToken)

    return .init(
      fileSystem: fileSystem,
      shell: shell,
      git: git,
      network: network,
      github: github,
      slack: slack,
      report: errorReporter
    )
  }
}
