// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// This struct has the responsability to notify developers of runtime errors during `FingerPullRequest` execution.
public struct ErrorReporter {
  /// Report a runtime error.
  public let report: (_ error: Error) throws -> Void

  /// Call as function the `report` property.
  public func callAsFunction(error: Error) throws {
    try self.report(error)
  }
}

extension ErrorReporter {
  /// Live implementation of an ErrorReporter that reports errors to a Slack channel.
  /// - Parameters:
  ///   - slack: The slack client.
  ///   - errorMessageComposer: The error message composer.
  ///   - channel: The Slack channel to use to report errors.
  /// - Returns: The live implementation
  public static func live(
    slack: Slack,
    composeErrorMessage: ErrorMessageComposer,
    channel: String
  ) -> Self {
    .init { error in
      let message = composeErrorMessage(for: error)
      try slack.send(message, channel)
    }
  }
}
