// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// FileSystem handler.
public struct FileSystem {
  /// Reads the content of a file from disk.
  ///
  /// # Parameters
  ///   - `file`: The file to be read.
  ///
  /// # Returns
  ///   The data read from the given file.
  public let read: (_ file: URL) throws -> Data

  /// Writes a Data object to a given file on disk.
  ///
  /// # Parameters
  ///   - `file`: The file to be written.
  ///   - `content`: The content to be written.
  public let write: (_ file: URL, _ content: Data) throws -> Void

  /// Removes the file or folder at the given path, if exists.
  ///
  /// # Parameters
  ///   - `path`: The path of the file or folder to be removed
  public let remove: (_ path: URL) throws -> Void

  /// Checks whether a file exists at the given URL.
  ///
  /// # Parameters
  ///   - `file`: The URL to be checked
  ///
  /// # Returns
  ///   `true` if the file exists on the filesystem.
  public let fileExists: (_ file: URL) -> Bool

  /// Returns whether the path passed as a parameter is a folder or not.
  ///
  /// # Parameters
  ///   - `path`: The path to check.
  ///
  /// # Returns
  ///   **true** if the path is a folder; **false** otherwise.
  public let isDirectory: (_ path: String) throws -> Bool

  /// Returns a string containing the path of the item pointed at by the symlink specified by 'path'
  ///
  /// # Parameters
  ///   - `path`: The path to resolve the symlink at.
  ///
  /// # Returns
  ///   The resolved path.
  public let destinationOfSymbolicLink: (_ path: String) throws -> String

  /// Returns an array of strings representing the filenames of the items in the specified directory and all its subdirectories
  /// recursively.
  ///
  /// # Parameters
  ///   - `path`: The path to start traversing at.
  ///
  /// # Returns
  ///   Filenames traversed recursiverly starting at the given path.
  public let subpathsOfDirectory: (_ path: String) throws -> [String]

  /// Returns whether the url passed as a parameter is a folder or not.
  ///
  /// # Parameters
  ///   - `url`: The url to check.
  ///
  /// # Returns
  ///   **true** if the url is a folder; **false** otherwise.
  public func isDirectory(url: URL) throws -> Bool {
    try self.isDirectory(url.expandingTilde.path)
  }

  /// Returns whether the path passed as a parameter exists and is a folder.
  ///
  /// # Parameters
  ///   - `path`: The path to check.
  ///
  /// # Returns
  ///   **true** if a file exists at the given path and it is a folder; **false** otherwise.
  public func existsAndDirectory(_ path: String) -> Bool {
    (try? self.isDirectory(path)) ?? false
  }

  /// Reads the content of a file, as UTF8 String, from disk.
  /// - Parameters:
  ///   - file: The file to be read.
  public func readString(_ file: URL) throws -> String {
    guard let string = String(data: try self.read(file), encoding: .utf8) else {
      throw Error.undecodableData(file)
    }
    return string
  }

  /// Writes a String to a given file on disk.
  /// - Parameters:
  ///   - file: The file to be written.
  ///   - content: The content to be written.
  public func write(_ file: URL, _ content: String) throws {
    try self.write(file, content.data(using: .utf8)!)
  }

  /// Live implementation
  public static func live(fileManager: FileManager = .default) -> Self {
    .init(
      read: { file in
        try Data(contentsOf: file)
      },
      write: { file, content in
        let expandedPath = file.expandingTilde
        try fileManager.createDirectory(
          at: expandedPath.deletingLastPathComponent(),
          withIntermediateDirectories: true
        )
        try content.write(to: expandedPath)
      },
      remove: { path in
        guard fileManager.fileExists(atPath: path.expandingTilde.path) else { return }
        try fileManager.removeItem(at: path.expandingTilde)
      },
      fileExists: { file in
        fileManager.fileExists(atPath: file.expandingTilde.path)
      },
      isDirectory: { path in
        var isDirectoryBool = ObjCBool(false)
        guard fileManager.fileExists(atPath: path, isDirectory: &isDirectoryBool) else {
          throw Error.pathDoesNotExist(path)
        }
        return isDirectoryBool.boolValue
      },
      destinationOfSymbolicLink: fileManager.destinationOfSymbolicLink(atPath:),
      subpathsOfDirectory: fileManager.subpathsOfDirectory(atPath:)
    )
  }
}

extension URL {
  fileprivate var expandingTilde: Self {
    return URL(fileURLWithPath: (self.relativePath as NSString).expandingTildeInPath)
  }
}

extension FileSystem {
  /// Error handling
  public enum Error: Equatable, LocalizedError {
    case undecodableData(URL)
    case pathDoesNotExist(String)

    public var errorDescription: String? {
      switch self {
      case .undecodableData(let url):
        return "Cannot decode content of \(url.relativePath) as UTF8"
      case .pathDoesNotExist(let path):
        return "Path \(path) does not exist"
      }
    }
  }
}
