// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// Git handler
public struct Git {
  /// Analyze `git log` in order to return the requested merged PR's numbers.
  /// The PR ids are returned in chronological order, not in the order on which they appear in the `git log` output.
  public let mergedPRs: (MergedPullRequestSelectionStrategy) throws -> [Int]

  /// Returns the remote url of the repo
  public let remoteURL: () throws -> URL?

  /// Returns the name of the current git branch.
  public let currentBranch: () throws -> String

  /// Returns the latest tag on the current git branch.
  public let latestTag: () throws -> String

  /// Adds files to git stage.
  public let add: (_ files: [URL]) throws -> Void

  /// Commits files in git stage.
  public let commit: (_ message: String) throws -> Void

  /// Pushes the current git head.
  public let push: (_ branch: String) throws -> Void

  /// Pulls the the provided branch.
  public let pull: (_ branch: String) throws -> Void
}

extension Git {
  /// Live implementation of Git
  /// - Parameter shell: The Shell
  static func live(shell: Shell) -> Self {
    .init(
      mergedPRs: { strategy in
        switch strategy {
        case .recently(limit: let limit):
          return try shell
            .executeOrThrow(
              "git log -n \(limit) --pretty=format:%s --grep=\"Merge pull request #\""
            ).output
            .split(separator: "\n")
            .map { String($0) }
            .compactMap { Self.parseMergePullRequest(header: $0) }
            .reversed()

        case .after(date: let date):
          let dateFormatter = DateFormatter()
          dateFormatter.dateFormat = "yyyy-MM-dd"
          let dateString = dateFormatter.string(from: date)

          return try shell
            .executeOrThrow(
              "git log --pretty=format:%s --grep=\"Merge pull request #\" --after=\"\(dateString)\""
            ).output
            .split(separator: "\n")
            .map { String($0) }
            .compactMap { Self.parseMergePullRequest(header: $0) }
            .reversed()

        case .until(commitMessage: let commitMessage):
          let commits = try shell
            .executeOrThrow(
              "git log --pretty=format:%s --grep=\"Merge pull request #\" --grep=\"\(commitMessage.escaped)\""
            ).output
            .split(separator: "\n")
            .map { String($0) }

          if commits.contains(commitMessage) {
            return commits
              .prefix(while: { $0 != commitMessage })
              .compactMap { Self.parseMergePullRequest(header: $0) }
              .reversed()
          } else {
            return commits
              .prefix(1)
              .compactMap { Self.parseMergePullRequest(header: $0) }
          }
        }
      },
      remoteURL: {
        let url = try shell.executeOrThrow("git config --get remote.origin.url").outputTrimmed
        return .init(string: url)
      },
      currentBranch: {
        let result = try shell.executeOrThrow("git branch --show-current")
        return result.outputTrimmed
      },
      latestTag: {
        let result = try shell.executeOrThrow("git describe --tags --abbrev=0")
        return result.outputTrimmed
      },
      add: { files in
        try shell.executeOrThrow("git add \(files.map { "\"\($0.relativePath)\"" }.joined(separator: " "))")
      },
      commit: { message in
        try shell.executeOrThrow("git commit -m \"\(message)\"")
      },
      push: { branch in
        /// This cannot be a simple `git push` because it would break in CircleCI.
        /// The reason is that CircleCI handles its `checkout` step in a weird way: it creates them and checks them out, so that
        /// they don't track their remote counterpart.
        /// Consequently, when pushing in the same repository that CircleCI is currently running in, a more explicit
        /// `git push origin HEAD:<branch_name>` must be used for it to work.
        /// This works exactly the same as a `git push` for a normally cloned repository, so it can be used in the general case.
        try shell.executeOrThrow("git push origin HEAD:\(branch)")
      },
      pull: { branch in
        /// This cannot be a simple `git pull` because it would break in CircleCI.
        /// The reason is that CircleCI handles its `checkout` step in a weird way: it creates them and checks them out, so that
        /// they don't track their remote counterpart.
        /// Consequently, when pushing in the same repository that CircleCI is currently running in, a more explicit
        /// `git push origin HEAD:<branch_name>` must be used for it to work.
        /// This works exactly the same as a `git pull` for a normally cloned repository, so it can be used in the general case.
        try shell.executeOrThrow("git pull origin HEAD:\(branch)")
      }
    )
  }
}

// When `RegexBuilder` will be available on MacOS we can delete the old implementation
// and just use the `RegexBuilder` one.
// I've added a `&& FALSE` to the precompiler check to make sure that we do not have a different
// behaviour on different MacOS versions.
#if canImport(RegexBuilder) && FALSE
  import RegexBuilder

  extension Git {
    /// The number of the pull request merged by this commit header. nil if this commit header is not a pull request merge commit.
    private static func parseMergePullRequest(header: String) -> Int? {
      let prNumberRegex = Regex {
        "Merge pull request #"
        Capture {
          OneOrMore(.digit)
        }
        " from "
        OneOrMore(.word)
      }.anchorsMatchLineEndings()

      return header.firstMatch(of: prNumberRegex)?.output.1
    }
  }
#else
  extension Git {
    /// The number of the pull request merged by this commit header. nil if this commit header is not a pull request merge commit.
    private static func parseMergePullRequest(header: String) -> Int? {
      guard let mergePullRequestRegEx = try? NSRegularExpression(pattern: "^Merge pull request #(\\d+) from .+$"),
            let mergeMatch = mergePullRequestRegEx.firstMatch(in: header, range: NSRange(location: 0, length: header.count)),
            let pullRequestNumberRange = Range(mergeMatch.range(at: 1), in: header),
            let pullRequestNumber = Int(header[pullRequestNumberRange])
      else {
        return nil
      }

      return pullRequestNumber
    }
  }
#endif

extension String {
  fileprivate var escaped: String {
    ["[", "]", "*", ".", "$"].reduce(self) { partialResult, character in
      partialResult.escape(character)
    }
  }

  private func escape(_ character: Character) -> String {
    let characterString = String(character)
    return self
      .replacingOccurrences(of: "\\\(characterString)", with: characterString)
      .replacingOccurrences(of: characterString, with: "\\\(characterString)")
  }
}
