// Copyright © Bending Spoons Group. All rights reserved.

import Foundation
import Logging

#if canImport(FoundationNetworking)
  import FoundationNetworking
#endif

// MARK: - Witness

/// Network witness
public struct Network {
  // swiftlint:disable identifier_name
  /// Performs a `GET` request
  public let _get: (
    _ address: URL,
    _ headers: [String: String],
    _ id: String
  ) throws -> Data

  /// Performs a `POST` request
  public let _post: (
    _ address: URL,
    _ headers: [String: String],
    _ body: Data,
    _ id: String
  ) throws -> Data

  /// Performs a `HEAD` request
  public let _head: (
    _ address: URL,
    _ headers: [String: String],
    _ id: String
  ) throws -> HTTPURLResponse

  /// Uploads a bulk of files
  public let _bulkUpload: (
    _ urlWithDestinations: [URL: URL],
    _ headers: (_ file: URL) -> [String: String],
    _ method: UploadHTTPMethod,
    _ id: String,
    _ logProgress: Bool
  ) throws -> Void

  // swiftlint:enable identifier_name

  /// Default init
  public init(
    get: @escaping (
      _ address: URL,
      _ headers: [String: String],
      _ id: String
    ) throws -> Data,
    post: @escaping (
      _ address: URL,
      _ headers: [String: String],
      _ body: Data,
      _ id: String
    ) throws -> Data,
    head: @escaping (
      _ address: URL,
      _ headers: [String: String],
      _ id: String
    ) throws -> HTTPURLResponse,
    bulkUpload: @escaping (
      _ urlWithDestinations: [URL: URL],
      _ headers: (_ file: URL) -> [String: String],
      _ method: UploadHTTPMethod,
      _ id: String,
      _ logProgress: Bool
    ) throws -> Void
  ) {
    self._get = get
    self._post = post
    self._head = head
    self._bulkUpload = bulkUpload
  }
}

// MARK: - Public interface

extension Network {
  /// Performs a `GET` request
  /// - parameters:
  ///   - address: the address of the request
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response body
  public func get(
    url address: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    id: String = UUID().uuidString
  ) throws -> Data {
    return try self._get(address.with(queryParams: queryParams), headers, id)
  }

  /// Performs a generic `POST` request
  /// - parameters:
  ///   - url: the address of the request
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - body: the body of the request (already encoded)
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response body
  public func post(
    url address: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    body: Data = Data(),
    id: String = UUID().uuidString
  ) throws -> Data {
    return try self._post(address.with(queryParams: queryParams), headers, body, id)
  }

  /// Performs a url form encoded `POST` request
  /// - parameters:
  ///   - url: the address of the request
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - form: the body of the request
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response body
  public func post<Payload: FormEncodable>(
    url address: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    form payload: Payload,
    id: String = UUID().uuidString
  ) throws -> Data {
    var headers = headers
    headers["Content-Type"] = "application/x-www-form-urlencoded; charset=utf-8"

    let data = payload.encoded.components.url?.query?.data(using: .utf8) ?? Data()

    return try self._post(address.with(queryParams: queryParams), headers, data, id)
  }

  /// Performs a `POST` request with an `Encodable` value
  /// - parameters:
  ///   - url: the address of the request
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - body: the body of the request
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response body
  public func post<Payload: Encodable>(
    url address: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    body payload: Payload,
    id: String = UUID().uuidString
  ) throws -> Data {
    var headers = headers
    headers["Content-Type"] = "application/json; charset=utf-8"

    let encoder = JSONEncoder()
    encoder.outputFormatting = [.sortedKeys, .prettyPrinted]
    let data = try encoder.encode(payload)

    return try self._post(address.with(queryParams: queryParams), headers, data, id)
  }

  /// Performs a `HEAD` request
  /// - parameters:
  ///   - url: the address of the request
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response headers
  public func head(
    url address: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    id: String = UUID().uuidString
  ) throws -> [String: String] {
    guard let allHeaderFields = try self._head(address.with(queryParams: queryParams), headers, id)
      .allHeaderFields as? [String: String] else { fatalError("Can't retrieve headers") }
    return allHeaderFields
  }

  /// Uploads a resource from the requested URL
  /// - parameters:
  ///   - localURL: the location on disk of the resource to upload
  ///   - destination: the address to which to upload the resource
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send
  ///   - method: the upload method to use
  ///   - id: a unique request id, used for logging
  /// - returns: a result containing the response headers
  public func upload(
    localURL: URL,
    destination: URL,
    queryParams: [String: String] = [:],
    headers: [String: String] = [:],
    method: UploadHTTPMethod = .POST,
    id: String = UUID().uuidString
  ) throws {
    try self._bulkUpload([localURL: destination.with(queryParams: queryParams)], { _ in headers }, method, id, false)
  }

  /// Uploads multiple resources to the given URLs
  /// - parameters:
  ///   - urlsWithDestination: the location on disk of the resourcess to upload, with the remote urls to which to upload them.
  ///   - queryParams: optional query parameters of the request
  ///   - headers: optional request headers to send, customizable for each uploaded file
  ///   - method: the upload method to use
  ///   - id: a unique request id, used for logging
  ///   - logProgress: whether the upload progress should be logged
  /// - returns: a result containing the response headers
  public func bulkUpload(
    urlsWithDestination: [URL: URL],
    queryParams: [String: String] = [:],
    headers: (_ file: URL) -> [String: String] = { _ in [:] },
    method: UploadHTTPMethod = .POST,
    id: String = UUID().uuidString,
    logProgress: Bool = false
  ) throws {
    let urlsWithQueryParams = try urlsWithDestination.mapValues { try $0.with(queryParams: queryParams) }
    try self._bulkUpload(urlsWithQueryParams, headers, method, id, logProgress)
  }
}

extension URL {
  fileprivate func with(queryParams: [String: String]) throws -> URL {
    guard var urlComponents = URLComponents(url: self, resolvingAgainstBaseURL: false) else {
      throw Network.Error.invalidURL(self.absoluteString)
    }

    if queryParams.isNotEmpty {
      urlComponents.queryItems = queryParams.components.queryItems
    }

    guard let url = urlComponents.url else {
      throw Network.Error.invalidURL(urlComponents.string ?? "")
    }

    return url
  }
}

extension Network {
  /// A network error
  public enum Error: LocalizedError, Equatable {
    /// The request failed.
    case requestFailed(request: URLRequest, code: Int, data: Data?)

    /// An error was returned
    case requestError(request: URLRequest, error: Swift.Error)

    /// The given string is not a valid URL
    case invalidURL(String)

    /// No response was received
    case noResponse(request: URLRequest)

    public var errorDescription: String? {
      switch self {
      case .requestFailed(let request, let code, let data):
        if let data = data, let string = String(data: data, encoding: .utf8) {
          return "\(Self.requestInfo(for: request)) failed with code: \(code). Data:\n\(string)"
        } else {
          return "\(Self.requestInfo(for: request)) failed with code: \(code)."
        }
      case .requestError(let request, let error):
        return "\(Self.requestInfo(for: request)) failed with error: \(error)"
      case .invalidURL(let url):
        return "Invalid URL: \(url)"
      case .noResponse(let request):
        return "No response received from \(Self.requestInfo(for: request))"
      }
    }

    static func requestInfo(for request: URLRequest) -> String {
      // swiftlint:disable:next force_unwrapping
      return "\(request.httpMethod ?? "GET") request to \(request.url!)"
    }

    public static func == (lhs: Network.Error, rhs: Network.Error) -> Bool {
      return lhs.errorDescription == rhs.errorDescription
    }
  }
}

extension Network {
  /// The available methods to perform the upload.
  public enum UploadHTTPMethod {
    /// A PUT upload.
    case PUT
    /// A POST upload.
    case POST

    var asHTTPMethod: String {
      switch self {
      case .POST:
        return "POST"
      case .PUT:
        return "PUT"
      }
    }
  }
}

// MARK: - FormEncodable

/// An entity that can be form encoded to be sent to a server
public protocol FormEncodable {
  /// the encoded representation of the entity
  var encoded: [String: String] { get }
}

extension Dictionary: FormEncodable where Key == String, Value == String {
  public var encoded: [String: String] {
    return self
  }
}

// MARK: - Helpers

extension Dictionary where Key == String, Value == String {
  fileprivate var components: URLComponents {
    var components = URLComponents()

    components.queryItems = self
      .map { URLQueryItem(name: $0, value: $1) }
      .sorted(by: { $0.name < $1.name }) // The sorting is done to keep the output predictable
    return components
  }
}

// MARK: - Live implementation

class BlockingRedirectDelegate: NSObject, URLSessionTaskDelegate {
  static let shared = BlockingRedirectDelegate()

  func urlSession(
    _: URLSession,
    task _: URLSessionTask,
    willPerformHTTPRedirection _: HTTPURLResponse,
    newRequest _: URLRequest,
    completionHandler: @escaping (URLRequest?) -> Void
  ) {
    completionHandler(nil)
  }
}

extension Network {
  /// Live implementation
  public static func live(
    shouldBlockRedirect: Bool = false,
    urlSessionBuilder: (URLSessionDelegate?)
      -> URLSession = { URLSession(configuration: .default, delegate: $0, delegateQueue: nil) },
    fileSystem: FileSystem
  ) -> Self {
    let logger = Logger(label: "Network")
    let urlSession = urlSessionBuilder(shouldBlockRedirect ? BlockingRedirectDelegate.shared : nil)

    return .init(
      get: { url, headers, id in
        return try Self.performSyncRequest(
          to: url,
          customize: { request in
            request.httpMethod = "GET"
            request.add(headers: headers)
          },
          logger: logger,
          urlSession: urlSession,
          id: id
        ).data
      },
      post: { url, headers, body, id in
        return try Self.performSyncRequest(
          to: url,
          customize: { request in
            request.httpMethod = "POST"
            request.add(headers: headers)
            request.httpBody = body
          },
          logger: logger,
          urlSession: urlSession,
          id: id
        ).data
      },
      head: { url, headers, id in
        return try Self.performSyncRequest(
          to: url,
          customize: { request in
            request.httpMethod = "HEAD"
            request.add(headers: headers)
          },
          logger: logger,
          urlSession: urlSession,
          id: id
        ).response
      },
      bulkUpload: { urlsWithDestination, headers, method, id, logProgress in
        return try Self.performSyncBulkUpload(
          urlWithDestinations: urlsWithDestination,
          customize: { request, file in
            request.httpMethod = method.asHTTPMethod
            request.add(headers: headers(file))
          },
          logger: logger,
          urlSession: urlSession,
          id: id,
          logProgress: logProgress
        )
      }
    )
  }

  fileprivate static func performSyncRequest(
    to url: URL,
    customize: (inout URLRequest) -> Void,
    logger: Logger,
    urlSession: URLSession,
    id: String
  ) throws -> (response: HTTPURLResponse, data: Data) {
    var request = URLRequest(url: url)
    customize(&request)

    let method = request.httpMethod ?? "GET"
    let semaphore = DispatchSemaphore(value: 0)
    var result: Result<(response: HTTPURLResponse, data: Data), Error>! // swiftlint:disable:this implicitly_unwrapped_optional

    logger.debug("[\(id)] \(method) request to \(url)")
    logger.trace("[\(id)] Headers: \(request.allHTTPHeaderFields ?? [:])")

    if let bodyData = request.httpBody,
       let bodyString = try? bodyData.prettyJsonEncoded() {
      logger.debug("[\(id)] Body: \(bodyString)")
    }

    let task: URLSessionDataTask = urlSession.dataTask(with: request) { data, response, error in
      if let response = response as? HTTPURLResponse,
         let data = data {
        if response.statusCode < 400 {
          logger
            .trace(
              "[\(id)] \(method) response: \(response) data: \(String(data: data, encoding: .utf8) ?? "CantDecode")"
            )
          result = .success((response: response, data: data))
        } else {
          result = .failure(Error.requestFailed(request: request, code: response.statusCode, data: data))
        }
      } else if let error = error {
        logger.debug("[\(id)] \(method) error: \(error)")
        result = .failure(.requestError(request: request, error: error))
      } else {
        result = .failure(.noResponse(request: request))
      }
      semaphore.signal()
    }
    task.resume()

    _ = semaphore.wait(timeout: .distantFuture)

    return try result.get()
  }

  fileprivate static func performSyncBulkUpload(
    urlWithDestinations: [URL: URL],
    customize: (inout URLRequest, _ file: URL) -> Void,
    logger: Logger,
    urlSession: URLSession,
    id: String,
    logProgress: Bool
  ) throws {
    let dispatchGroup = DispatchGroup()
    var results: [URL: Result<Void, Error>] = [:]

    var index = 0
    for (localUrl, destination) in urlWithDestinations {
      index += 1
      let id = "\(id)-\(index)"
      dispatchGroup.enter()
      var request = URLRequest(url: destination)
      customize(&request, localUrl)

      logger.debug("[\(id)] Upload request of file \(localUrl) to \(destination)")

      let task = urlSession.uploadTask(with: request, fromFile: localUrl) { data, response, error in
        defer { dispatchGroup.leave() }

        guard let response = response as? HTTPURLResponse else {
          logger.debug("[\(id)] Upload error: NO RESPONSE")
          results[localUrl] = .failure(.noResponse(request: request))
          return
        }

        guard response.statusCode < 400 else {
          logger.debug("[\(id)] Upload error: Server responded with \(response.statusCode)")
          results[localUrl] = .failure(.requestFailed(request: request, code: response.statusCode, data: data))
          return
        }

        if let error = error {
          logger.debug("[\(id)] Upload error: \(error)")
          results[localUrl] = .failure(.requestError(request: request, error: error))
          return
        }

        logger.debug("[\(id)] Upload response: \(response.statusCode)")
        results[localUrl] = .success(())

        if logProgress {
          logProgressForBulkOperation(
            completedOperations: results.count,
            totalOperations: urlWithDestinations.count,
            operationName: "uploads",
            logger: logger
          )
        }
      }

      task.resume()
    }

    _ = dispatchGroup.wait(timeout: .distantFuture)

    _ = try results.map(\.value).getSuccessesOrThrow()
  }
}

// MARK: - Helpers

extension URLRequest {
  mutating func add(headers: [String: String]) {
    for (key, value) in headers {
      /// Capitalization is needed to keep behaviour consistent with Linux (as there all the headers
      /// are capitalized by default)
      self.addValue(value, forHTTPHeaderField: key.capitalized)
    }
  }
}

/// If the total number of operations is more than one, then it logs a message with the current percentage at the following steps:
/// - after the first operation
/// - after every operation for which the total progress reaches a whole multiple of 10 (for example, at the first operation
///   that makes the progress go over 10%, 20%, etc).
/// - after the last operation
///
/// - Parameters:
///   - completedOperations: the count of completed operations
///   - totalOperations: the count of total operations
///   - operationName: the name of the operation, to be used in the log message
///   - logger: the logger
private func logProgressForBulkOperation(
  completedOperations: Int,
  totalOperations: Int,
  operationName: String,
  logger: Logger
) {
  let previousProgressPercentage = Int(100 * Double(completedOperations - 1) / Double(totalOperations))
  let currentProgressPercentage = Int(100 * Double(completedOperations) / Double(totalOperations))

  // This is true if the current progress is exactly one operation above a whole multiple of 10, in percentage.
  // For example, if `previousProgressPercentage == 39` and `currentProgressPercentage == 41`.
  let isFirstOperationAfterMultipleOfTen = previousProgressPercentage / 10 < currentProgressPercentage / 10
  let isFirstPrint = completedOperations == 1
  let isLastPrint = completedOperations == totalOperations

  if isFirstPrint || isFirstOperationAfterMultipleOfTen || isLastPrint {
    logger.info("\(currentProgressPercentage)% \(operationName) completed")
  }
}
