// Copyright © Bending Spoons Group. All rights reserved.

import Foundation
import Logging

// MARK: Interface

/// Helper to execute shell commands.
public struct Shell {
  // swiftlint:disable:next identifier_name
  let _execute: (
    _ command: String,
    _ environment: [String: EnvironmentValue],
    _ outputLogLevel: Logger.Level,
    _ errorLogLevel: Logger.Level,
    _ logFile: URL?,
    _ logsSensitiveInformation: Bool
  ) -> Result

  /// Synchronously executes a shell command in the current folder and returns its result.
  ///
  /// - Parameters:
  ///   - command: The command to be executed.
  ///   - outputLogLevel: The log level to be used for command standard output.
  ///   - errorLogLevel: The log level to be used for command standard error.
  ///   - logFile: The log file where the standard output and error should be redirected.
  ///   - logsSensitiveInformation: Whether the command might log sensitive information.
  ///
  /// # Returns
  ///   The result containing the status code, the standard output, and the standard error.
  @discardableResult
  public func execute(
    _ command: String,
    environment: [String: EnvironmentValue] = [:],
    outputLogLevel: Logger.Level = .debug,
    errorLogLevel: Logger.Level = .debug,
    logFile: URL? = nil,
    logsSensitiveInformation: Bool = false
  ) -> Result {
    return self._execute(command, environment, outputLogLevel, errorLogLevel, logFile, logsSensitiveInformation)
  }

  /// Throwing version of `execute(:)`.
  /// It executes the command, but throws if the command fails
  /// - Parameters:
  ///   - command: The command to be executed.
  ///   - environment: The environment to be setup for the command.
  ///   - outputLogLevel: The log level to be used for command standard output.
  ///   - errorLogLevel: The log level to be used for command standard error.
  ///   - logFile: The log file where the standard output and error should be redirected.
  ///   - logsSensitiveInformation: Whether the command might log sensitive information.
  /// - Returns: The result containing the status code, the standard output, and the standard error.
  @discardableResult
  public func executeOrThrow(
    _ command: String,
    environment: [String: EnvironmentValue] = [:],
    outputLogLevel: Logger.Level = .debug,
    errorLogLevel: Logger.Level = .debug,
    logFile: URL? = nil,
    logsSensitiveInformation: Bool = false
  ) throws -> Result {
    let result = self._execute(command, environment, outputLogLevel, errorLogLevel, logFile, logsSensitiveInformation)

    guard !result.isFailed else {
      throw Error.commandFailed(command: command, errorMessage: result.resultMessage)
    }

    return result
  }
}

// MARK: Types

extension Shell {
  /// The result of a shell command execution.
  public struct Result: Equatable {
    /// Command result.
    public let status: Int32

    /// Command standard error.
    public let output: String

    /// Command standard output.
    public let error: String
  }

  /// An error that may occur while running a shell command
  public enum Error: LocalizedError, Equatable {
    /// The command failed with the given error message
    case commandFailed(command: String, errorMessage: String)

    public var errorDescription: String? {
      switch self {
      case .commandFailed(let command, let errorMessage):
        return "Command `\(command)` failed with message: \(errorMessage)"
      }
    }
  }

  /// A shell environment value.
  public struct EnvironmentValue: Equatable, ExpressibleByBooleanLiteral, ExpressibleByIntegerLiteral,
    ExpressibleByStringLiteral {
    /// The actual value.
    public let value: String

    /// Whether the value contains sensitive information
    public let isSensitive: Bool

    /// Default init
    public init(value: String, isSensitive: Bool) {
      self.value = value
      self.isSensitive = isSensitive
    }

    public init(stringLiteral: String) {
      self.init(value: stringLiteral, isSensitive: false)
    }

    public init(booleanLiteral: Bool) {
      self.init(value: booleanLiteral ? "1" : "0", isSensitive: false)
    }

    public init(integerLiteral: Int) {
      self.init(value: "\(integerLiteral)", isSensitive: false)
    }
  }
}

// MARK: Implementation

extension Shell {
  /// A bash shell.
  /// - Parameter folder: The folder in which git commands should be executed. If nil, the current folder is used.
  public static func bash(in folder: URL? = nil) -> Shell {
    return .bash(
      in: folder,
      processBuilder: { Process() },
      outputPipeBuilder: { Pipe() },
      errorPipeBuilder: { Pipe() },
      log: { Self.log(commandID: $0, content: $1, label: $2, logger: $3, level: $4, isSensitive: $5) }
    )
  }

  static func bash(
    in folder: URL? = nil,
    processBuilder: @escaping () -> Process,
    outputPipeBuilder: @escaping () -> Pipe,
    errorPipeBuilder: @escaping () -> Pipe,
    log: @escaping (Int, String, String?, Logger, Logger.Level, Bool) -> Void
  ) -> Shell {
    return .init(_execute: { command, environment, outputLogLevel, errorLogLevel, logFile, mightContainSensitiveInformation in
      let logger = Logger(label: "Shell")
      let process = processBuilder()
      process.launchPath = "/bin/bash"
      let commandWithTee: String
      if let logFile = logFile {
        commandWithTee = "\(command) 2>&1 | tee -a \(logFile.path)"
      } else {
        commandWithTee = command
      }
      let commandWithEnvironment = (environment.map { "\($0.key)=\($0.value.value)" }.sorted() + [commandWithTee])
      process.arguments = ["-eo", "pipefail", "-c", commandWithEnvironment.joined(separator: " ")]
      process.currentDirectoryURL = folder
      let outputPipe = outputPipeBuilder()
      process.standardOutput = outputPipe
      let errorPipe = errorPipeBuilder()
      process.standardError = errorPipe

      let commandID = folder.hashValue ^ command.hashValue
      let environmentArguments: String
      if environment.isEmpty {
        environmentArguments = ""
      } else {
        environmentArguments = environment
          .map { key, value in "\(key)=\(value.isSensitive ? "******" : "\(value.value)")" }
          .sorted()
          .joined(separator: " ") + " "
      }
      let commandLog = "Starting `\(environmentArguments)\(command)`"
      log(commandID, commandLog, nil, logger, max(outputLogLevel, errorLogLevel), false)

      process.launch()

      // We must call `waitTerminationAndReadData` concurrently on both pipe, because of large
      // output or error the Pipe gets filled and the execution blocks.
      let pipeOutputs = [outputPipe, errorPipe].concurrentMap {
        $0.waitTerminationAndReadData
      }
      let output = pipeOutputs[0]
      let error = pipeOutputs[1]

      log(commandID, output, "STDOUT", logger, outputLogLevel, mightContainSensitiveInformation)
      log(commandID, error, "STDERR", logger, errorLogLevel, mightContainSensitiveInformation)

      process.waitUntilExit()

      let result = Result(status: process.terminationStatus, output: output, error: error)
      logger.debug("[\(commandID)] Finished `\(command)`. \(result.resultMessage)")
      return result
    })
  }

  fileprivate static func log(
    commandID: Int,
    content: String,
    label: String?,
    logger: Logger,
    level: Logger.Level,
    isSensitive: Bool
  ) {
    #if !DEBUG
      // Only in production, check if the log might contain sensitive information.
      guard !isSensitive else {
        // If that's the case, don't print anything.
        return
      }
    #endif

    let prefix = "[\(commandID)] \(label.map { "[\($0)] " } ?? "")"
    for line in content.split(separator: "\n") {
      logger.log(level: level, "\(prefix)\(line)")
    }
  }
}

extension Pipe {
  /// Returns the content of the Pipe as a String
  fileprivate var waitTerminationAndReadData: String {
    return String(data: self.fileHandleForReading.readDataToEndOfFile(), encoding: String.Encoding.utf8) ?? ""
  }
}

// MARK: - Utilities

extension Shell.Result {
  /// Standard output trimmed of all whitespace
  public var outputTrimmed: String {
    return self.output.trimmingCharacters(in: .whitespacesAndNewlines)
  }

  /// Standard error trimmed of all whitespace
  public var errorTrimmed: String {
    return self.error.trimmingCharacters(in: .whitespacesAndNewlines)
  }

  /// The result message of the command. Formatted as "code" for successful commands and "code: message" for failed ones.
  public var resultMessage: String {
    if self.isSuccess {
      return String(self.status)
    } else {
      let errorMessage: String
      if !self.errorTrimmed.isEmpty {
        errorMessage = self.errorTrimmed
      } else if !self.outputTrimmed.isEmpty {
        errorMessage = self.outputTrimmed
      } else {
        errorMessage = "Unknown Error"
      }
      return "\(self.status): \(errorMessage)"
    }
  }

  /// Whether the command is failed or not
  public var isFailed: Bool {
    return self.status != 0
  }

  /// Whether the command is success or not
  public var isSuccess: Bool {
    return !self.isFailed
  }
}
