// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// This struct has the responsability to sent messages to slack.
public struct Slack {
  /// Send a message to a particular slack channel.
  public var send: (_ message: String, _ channel: String) throws -> Void
}

extension Slack {
  /// Returns the live implementation of the Slack client
  /// - Parameters:
  ///   - network: The network dependency.
  ///   - slackAccessToken: The Slack access token.
  /// - Returns: The live implementation of the Slack client
  public static func live(
    network: Network,
    slackAccessToken: String
  ) -> Self {
    .init { message, channel in
      let body = Body(text: message, channel: channel)
      _ = try network.post(
        url: URL(string: "https://slack.com/api/chat.postMessage")!,
        headers: [
          "Authorization": "Bearer \(slackAccessToken)",
        ],
        body: body
      )
    }
  }
}

extension Slack {
  fileprivate struct Body: Encodable {
    let text: String
    let channel: String
    let linkNames = true

    enum CodingKeys: String, CodingKey {
      case text
      case channel
      case linkNames = "link_names"
    }
  }
}
