// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// This is the main object that allows to write scripts based on the PullRequest of a Repository.
public struct FingerPullRequest {
  /// The environment
  public let environment: Environment

  /// Default initializer
  /// - Parameter environment: The environment used
  public init(environment: Environment) {
    self.environment = environment
  }

  /// Execute the `handler` logic on the PRs selected by the provided `strategy`.
  /// - Parameters:
  ///   - strategy: Defines on which PRs you want to execute the `handler` logic.
  ///   - handler: Allow the user to perform any kind of work on the provided PR,
  ///   The handler once per PR with the PRs in chronological order.
  public func run(
    on strategy: PullRequestSelectionStrategy,
    handler: (PullRequest) throws -> Void
  ) throws {
    do {
      let prIDs: [Int]

      switch strategy {
      case .merged(let mergedStrategy):
        prIDs = try self.environment.git.mergedPRs(mergedStrategy)

      case .select(let prNumbers):
        prIDs = prNumbers
      }

      try prIDs.forEach { id in
        let pr = try self.environment.github.fetchPullRequest(id)
        try handler(pr)
      }
    } catch Error.dangerValidation(let error) {
      throw error
    } catch {
      try? self.environment.report(error: error)
      throw error
    }
  }
}

extension FingerPullRequest {
  /// The `FingerPullRequest` Live initializer
  /// - Parameters:
  ///   - errorReporterSlackChannel: The Slack channel where the runtime errors are notified.
  ///   - errorResponsiblePeople: The list of Slack handles that will be tagged in case of an error.
  ///   - githubAuthToken: The Github access token, by default it uses the `GITHUB_TOKEN` env var.
  ///   - slackAccessToken: The Slack access token, by default it uses the `SLACK_ACCESS_TOKEN` env var.
  ///   - repositoryName: The repositoryName, if not provided is extracted automatically from Git.
  /// - Returns: The `FingerPullRequest` instance.
  public static func live(
    errorReporterSlackChannel: String,
    errorResponsiblePeople: [String] = ["channel"],
    githubAuthToken: String! = ProcessInfo.processInfo.environment["GITHUB_TOKEN"],
    slackAccessToken: String! = ProcessInfo.processInfo.environment["SLACK_ACCESS_TOKEN"],
    repositoryName: String? = nil
  ) throws -> Self {
    guard githubAuthToken != nil else { throw Error.missingGithubAuthToken }
    guard slackAccessToken != nil else { throw Error.missingSlackAccessToken }

    return .init(
      environment: .live(
        errorReporterSlackChannel: errorReporterSlackChannel,
        errorResponsiblePeople: errorResponsiblePeople,
        githubAuthToken: githubAuthToken,
        slackAccessToken: slackAccessToken,
        repositoryName: repositoryName
      )
    )
  }
}

extension FingerPullRequest {
  /// Errors related to missing required auth tokens.
  public enum Error: Swift.Error, LocalizedError {
    /// Missing the Github access token
    case missingGithubAuthToken
    /// Missing the Slack access token
    case missingSlackAccessToken
    /// An error due to danger validation of a PR.
    /// In Danger we accept that a PR may be invalid and we don't need to notify on slack fro this.
    case dangerValidation(Swift.Error)

    /// Error description.
    public var errorDescription: String? {
      switch self {
      case .missingGithubAuthToken:
        return "Missing 'githubAuthToken', set it in 'GITHUB_TOKEN' env var."
      case .missingSlackAccessToken:
        return "Missing 'slackAccessToken', set it in 'SLACK_ACCESS_TOKEN' env var."
      case .dangerValidation(let error):
        return error.localizedDescription
      }
    }
  }
}
