// Copyright © Bending Spoons Group. All rights reserved.

import Foundation
import Markdown

/// A markdown document
/// Once created you can query the document in order to easily extract informations regarding
/// the value of different `MarkdownDocumentElements`s.
/// example:
/// ```
/// static let changelogText = MarkdownElements.TextInput("Changelog Entry:", description: "Insert here the changelog")
/// static let updateChangelog = MarkdownElements.Checkbox("Should update the changelog")
///
/// let markdown = """
///  - [x] Should update the changelog
///
///  ### Changelog Entry
///  > Insert here the changelog
///  - My custom changelog text.
/// """
///
/// let document = MarkdownDocument(parsing: markdown)
/// print(document.getValue(updateChangelog)) // Prints: true
/// print(document.getValue(changelogText)) // Prints: `My custom changelog text.`
/// ```
public struct MarkdownDocument {
  let document: Document

  /// Default init
  public init(parsing text: String) {
    self.document = Document(parsing: text)
  }

  /// Extract and return the value of a `MarkdownElement`.
  public func getValue<Element: MarkdownDocumentElement>(of element: Element) -> Element.Walker.Value {
    var walker = element.walker
    walker.visit(self.document)
    return walker.value
  }
}

// MARK: - MarkdownDocument Walkers

/// Protocol that describes a `MarkupWalker` with a return value.
public protocol MarkdownDocumentWalker: MarkupWalker {
  associatedtype Value
  /// The value returned by the walker.
  var value: Value { get }
}

/// Describes a markdown element that can be extracted from a `MarkdownDocument`.
public protocol MarkdownDocumentElement {
  /// The `MarkdownDocumentWalker` able to parse this element.
  associatedtype Walker: MarkdownDocumentWalker

  /// An instance of `Walker`.
  var walker: Walker { get }
}

extension MarkdownElements.Checkbox: MarkdownDocumentElement {
  /// A walker able to parse this element.
  public var walker: Walker {
    .init(self)
  }

  /// Walker that can parse a `Checkbox`
  public struct Walker: MarkdownDocumentWalker {
    let checkbox: MarkdownElements.Checkbox
    /// Return value
    public var value: Bool?
    // swiftlint:disable:previous discouraged_optional_boolean

    init(_ checkbox: MarkdownElements.Checkbox) {
      self.checkbox = checkbox
    }

    /// Visitor method
    public mutating func visitListItem(_ listItem: ListItem) {
      if let check = listItem.checkbox,
         self.markdown(for: listItem).contains(self.checkbox.markdownTitle) {
        self.value = check == .checked
        return
      }
      self.descendInto(listItem)
    }
  }
}

extension MarkdownElements.TextInput: MarkdownDocumentElement {
  /// A walker able to parse this element.
  public var walker: Walker {
    .init(self)
  }

  /// Walker that can parse a `TextInput`
  public struct Walker: MarkdownDocumentWalker {
    enum Status: Equatable {
      case searchingTitle
      case searchingDescription
      case searchingInput
      case done
    }

    /// Return value
    public var value: String?
    let textInput: MarkdownElements.TextInput
    private var status: Status = .searchingTitle

    init(_ textInput: MarkdownElements.TextInput) {
      self.textInput = textInput
    }

    /// Visitor method
    public mutating func visitHeading(_ heading: Heading) {
      if heading.level == self.textInput.heading.rawValue,
         self.markdown(for: heading) == self.textInput.markdownTitle {
        self.status = .searchingDescription
      } else {
        self.status = .searchingTitle
        self.descendInto(heading)
      }
    }

    /// Visitor method
    public mutating func visitBlockQuote(_ blockQuote: BlockQuote) {
      if self.status == .searchingDescription {
        self.status = .searchingInput
      } else {
        self.status = .searchingTitle
        self.descendInto(blockQuote)
      }
    }

    /// Visitor method
    public mutating func visitListItem(_ listItem: ListItem) {
      if self.status == .searchingInput {
        let textInserted = self.markdown(for: listItem)
        self.value = textInserted
        self.status = .done
      } else {
        self.descendInto(listItem)
      }
    }
  }
}

// MARK: - Utils

extension MarkupWalker {
  /// Returns the text representation of a markdown element
  fileprivate func markdown<Element: BlockMarkup>(for element: Element) -> String {
    let document = Document(element)
    var formatter = MarkupFormatter()
    formatter.visitDocument(document)
    return formatter.result
  }
}
