// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

/// Describe a reference to markdown elements
public enum MarkdownElements {}

// MARK: - Checkbox

extension MarkdownElements {
  /// A reference to the markdown element representing a checkbox that can be checked/unchecked in a given markdown file.
  /// This struct is used to point to an existing checkbox item in a markdown document that can be referenced
  /// using its title and eventually its description.
  public struct Checkbox: MarkdownRepresentable {
    /// Checkbox title.
    public let title: String

    /// Optional notes after the title to provide more context
    public let description: String

    /// Default init
    /// - Parameter title: Title of the checkbox
    /// - description: Text after the title used to provide more context ot usage hints
    public init(_ title: String, description: String = "") {
      self.title = title
      self.description = description
    }
  }
}

// MARK: - TextInput

extension MarkdownElements {
  /// A reference the markdown element representing a text input that can be filled in a given markdown file.
  /// This struct is used to point to an existing text input item in a markdown document that can be referenced
  /// using its title and eventually its description.
  public struct TextInput: MarkdownRepresentable {
    /// The title.
    public let title: String
    /// The description.
    public let description: String
    /// The kind of heading
    public let heading: HeadingKind

    /// Default init
    /// - Parameters:
    ///   - title: The title above the inserted text
    ///   - description: The description of the text input
    ///   - heading: The kind of heading to use for rendering
    public init(_ title: String, description: String, heading: HeadingKind = .h3) {
      self.title = title
      self.description = description
      self.heading = heading
    }
  }
}

/// The kind of markdown Heading
public enum HeadingKind: Int {
  case h1 = 1
  case h2
  case h3
  case h4
  case h5
  case h6

  var markdown: String {
    String(repeating: "#", count: self.rawValue)
  }
}
