// Copyright © Bending Spoons Group. All rights reserved.

import Foundation
import Markdown

/// This protocol represent an element that can know how to represented itself in Markdown.
public protocol MarkdownRepresentable {
  /// Textual representation of the markdown element.
  var markdown: String { get }
}

// MARK: - Checkbox

extension MarkdownElements.Checkbox {
  var markdownCheckbox: String {
    "- [ ]"
  }

  var markdownTitle: String {
    self.title.markdownRepresentation
  }

  var markdownDescription: String {
    guard self.description.isEmpty == false else { return "" }
    return "*\(description.markdownRepresentation.trimmingCharacters(in: .whitespaces))*"
  }

  /// Mardown text representation of this item.
  public var markdown: String {
    String(
      [
        self.markdownCheckbox,
        self.markdownTitle,
        self.markdownDescription,
      ].compactMap { $0 }
        .joined(separator: " ")
    )
  }
}

// MARK: - TextInput

extension MarkdownElements.TextInput {
  var markdownTitle: String {
    "\(self.heading.markdown) \(self.title)".markdownRepresentation
  }

  var markdownDescription: String {
    "> \(self.description)".markdownRepresentation
  }

  var markdownTextField: String {
    "- "
  }

  /// Mardown text representation of this item.
  public var markdown: String {
    String([
      self.markdownTitle,
      self.markdownDescription,
      self.markdownTextField,
    ].joined(separator: "\n"))
  }
}

extension String {
  /// Encode and decode a string to markdown.
  /// This is used because when parsed some symbol may be replaced for example the
  /// single quotation mark are replaced with the apostrophe.
  fileprivate var markdownRepresentation: String {
    let document = Document(parsing: self)
    var formatter = MarkupFormatter()
    formatter.visitDocument(document)
    return formatter.result
  }
}
