// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

extension FingerPullRequest {
  /// This method append one entry for every PR merged to the `unreleased` section in `CHANGELOG.md` file.
  /// You need to generate the new changelog entry for every PR merged using the provided `handler`.
  /// This method fetches all the PRs that has been merged until the provided commit message is present in the git history.
  /// If it's the first time that this tool runs it will execute just on the most recent PR merge
  /// because the target commit message is not present in the git history.
  /// - Parameters:
  ///   - file: URL to the `CHANGELOG.md` file, by default is `../CHANGELOG.md`
  ///   - commitMessage: The commit message used to commit the changelog update, default: `"docs: changelog.sh update [skip ci]"`
  ///   - handler: This closure must return the changelog entry given a specific PR.
  public func updateChangelog(
    file: URL = .init(fileURLWithPath: "../CHANGELOG.md"),
    commitMessage: String = "docs: changelog.sh update [skip ci]",
    changelogEntry: (PullRequest) throws -> String
  ) throws {
    var newEntries: [String] = []
    try self.run(on: .merged(.until(commitMessage: commitMessage))) { pr in
      let newEntry = try changelogEntry(pr)
      newEntries.append(newEntry)
    }

    guard !newEntries.isEmpty else {
      print("No new changelog entries to add.")
      return
    }

    do {
      let currentBranch = try self.environment.git.currentBranch()
      #if DEBUG
        try self.environment.git.pull("origin/" + currentBranch)
      #else
        try self.environment.git.pull(currentBranch)
      #endif
      let changelog = try self.environment.fileSystem.readString(file)
      let newChangelog = try changelog.addingToUnreleasedSection(entries: newEntries)
      try self.environment.fileSystem.write(file, newChangelog)
      try self.environment.git.add([file])
      try self.environment.git.commit(commitMessage)
      try self.environment.git.push(currentBranch)
    } catch {
      try? self.environment.report(error: error)
      throw error
    }
  }
}

extension String {
  fileprivate func addingToUnreleasedSection(entries: [String]) throws -> String {
    var changelog = self
    let unreleased = "## unreleased\n\n"

    // Add 'Unreleased' section if needed.
    if changelog.range(of: unreleased) == nil {
      let header = "# Changelog"
      guard let range = changelog.range(of: header) else { throw UpdateChangelogError.missingHeader }
      changelog = changelog.replacingCharacters(in: range, with: header + "\n\n" + unreleased)
    }

    guard let range = changelog.range(of: unreleased) else { throw UpdateChangelogError.missingUnreleasedSection }
    let formattedEntries = entries
      .map { "- " + $0 + "\n" }
      .joined()

    return changelog.replacingCharacters(in: range, with: unreleased + formattedEntries)
  }
}

/// Update Changelog specific errors
public enum UpdateChangelogError: Error {
  case missingHeader
  case missingUnreleasedSection
}
