// Copyright © Bending Spoons Group. All rights reserved.

import Foundation

extension Collection where Element: Hashable {
  /// Returns the collection as Set
  public var asSet: Set<Element> {
    return Set(self)
  }
}

/// Namespace of helpers to work with `Collections`
public enum CollectionHelpers {
  /// Handler to orchestrate the concurrent operations on the collections
  public struct ConcurrentOrchestrator {
    /// Submits a single block to the dispatch queue and causes the block to be executed the specified number of times.
    ///
    /// This method implements an efficient parallel for-loop. The dispatch queue executes the submitted block the specified
    /// number of times and waits for all iterations to complete before returning. If the target queue is a concurrent queue, the
    /// blocks run in parallel and must therefore be reentrant-safe.
    ///
    /// # Parameters
    ///
    /// - `iterations`:  The number of times to execute the block. Higher iteration values give the system the ability to balance
    ///   more efficiently across multiple cores.
    /// - `execute`: The block to execute in parallel. This block has no return value and takes the following parameter:
    ///   - `iteration`: The current iteration index.
    let concurrentPerform: (_ iterations: Int, _ execute: (Int) -> Void) -> Void
  }
}

extension CollectionHelpers.ConcurrentOrchestrator {
  /// Default implementation of a `ConcurrentOrchestrator`, based on the `DispatchQueue`.
  public static var dispatchQueueOrchestrator: Self {
    return .init(concurrentPerform: DispatchQueue.concurrentPerform(iterations:execute:))
  }

  /// Default implementation of a `ConcurrentOrchestrator`, based on the `DispatchQueue` with a limit on concurrent jobs count.
  public static func dispatchQueueOrchestrator(maxConcurrent: Int?) -> Self {
    guard let maxConcurrent = maxConcurrent else {
      return Self.dispatchQueueOrchestrator
    }

    let semaphore = DispatchSemaphore(value: maxConcurrent)
    return .init(concurrentPerform: { iterations, execute in
      DispatchQueue.concurrentPerform(
        iterations: iterations,
        execute: { iteration in
          semaphore.wait()
          execute(iteration)
          semaphore.signal()
        }
      )
    })
  }
}

extension Collection {
  /// Maps the elements of a collection in parallel.
  /// It ensures thread safety as long as the collection itself is not captured and accessed within the closure.
  @discardableResult
  public func concurrentMap<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Element) -> R
  ) -> [R] {
    let throwingTransform: (Element) throws -> R = transform
    return try! self.concurrentMap(concurrentOrchestrator, throwingTransform) // swiftlint:disable:this force_try
  }

  /// Maps the elements of a collection in parallel.
  /// It ensures thread safety as long as the collection itself is not captured and accessed within the closure.
  /// _Note_: it rethrows one of the first errors being thrown by an iteration of the transform function
  @discardableResult
  public func concurrentMap<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Element) throws -> R
  ) throws -> [R] {
    let results = ThreadSafe([R?](repeating: nil, count: self.count))
    let firstError: ThreadSafe<Error?> = .init(nil)

    concurrentOrchestrator.concurrentPerform(self.count) { i in
      guard firstError.value == nil else {
        // Early exit in case some other iteration has failed
        return
      }

      let result: R
      do {
        result = try transform(self[self.index(self.startIndex, offsetBy: i)])
      } catch {
        firstError.atomically { $0 = error }
        return
      }

      results.atomically { $0[i] = result }
    }

    if let error = firstError.value {
      throw error
    }

    return results.value.map { $0! } // swiftlint:disable:this force_unwrapping
  }
}

extension Collection {
  /// Performs a compactMap concurrently. See `concurrentMap`
  public func concurrentCompactMap<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Element) -> R?
  ) -> [R] {
    return self.concurrentMap(concurrentOrchestrator, transform)
      .compactMap { $0 }
  }

  /// Throwable version of `concurrentCompactMap`
  public func concurrentCompactMap<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Element) throws -> R?
  ) throws -> [R] {
    return try self.concurrentMap(concurrentOrchestrator, transform)
      .compactMap { $0 }
  }
}

extension Dictionary {
  /// Performs a mapValues concurrently. See `concurrentMap`
  public func concurrentMapValues<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Value) -> R
  ) -> [Key: R] {
    return .init(uniqueKeysWithValues: self.concurrentMap(concurrentOrchestrator) { key, value in (key, transform(value)) })
  }

  /// Performs a mapValues concurrently. See `concurrentMap`
  public func concurrentMapValues<R>(
    _ concurrentOrchestrator: CollectionHelpers.ConcurrentOrchestrator = .dispatchQueueOrchestrator,
    _ transform: @escaping (Value) throws -> R
  ) throws -> [Key: R] {
    return .init(uniqueKeysWithValues: try self.concurrentMap(concurrentOrchestrator) { key, value in
      (key, try transform(value))
    })
  }
}

extension Collection where Index == Int {
  /// Splits collection into chunks of size <= `size`
  /// - Parameter size: size of a single chunk
  /// - Returns: array of chunks
  public func chunked(size: Int) -> [[Element]] {
    let elementCount = count
    return stride(from: 0, to: count, by: size).map {
      Array(self[$0 ..< Swift.min($0 + size, elementCount)])
    }
  }
}

extension Dictionary {
  /// Transforms the keys of a given dictionary by applying the `transform` function.
  ///
  /// _Note_: it assumes that the mapping is 1 to 1, otherwise it throws a fatalError.
  public func mapKeys<T: Hashable>(_ transform: (Key) throws -> T) rethrows -> [T: Value] {
    return try self.reduce(into: [:]) { result, dictionary in
      let newKey = try transform(dictionary.key)

      guard result[newKey] == nil else {
        fatalError("Key mapping is not 1 to 1")
      }

      result[newKey] = dictionary.value
    }
  }

  /// Returns the result of the merge of two dictionaries, and throws a fatal error in case of duplicated keys
  public func mergeOrFail(_ other: Self) -> Self {
    return self.merging(other, uniquingKeysWith: { _, _ in
      let duplicatedKeys = self.keys.asSet.intersection(other.keys.asSet)
      fatalError("Duplicated keys during mergeOrFail: \(duplicatedKeys)")
    })
  }
}

extension Collection {
  /// Returns `true` if the collection is not empty
  public var isNotEmpty: Bool {
    return !self.isEmpty
  }
}
