// Copyright © Bending Spoons Group. All rights reserved.

import XCTest
@testable import FingerPullRequest

class GitTests: XCTestCase {
  var shell: Shell!
  var git: Git!

  override func setUpWithError() throws {
    try super.setUpWithError()
    self.shell = .bash(in: FileManager.default.temporaryDirectory)
    self.git = .live(shell: self.shell)
    try self.createGitRepositoryForTesting()
  }

  override func tearDownWithError() throws {
    try super.tearDownWithError()
    _ = try self.shell.executeOrThrow("rm -rf .git")
  }

  func testMergedPrs_mergedRecently() throws {
    let prs = try self.git.mergedPRs(.recently(limit: 100))
    XCTAssertEqual(prs, [5102, 5105, 5065, 5090, 5066, 5094])
  }

  func testMergedPrs_mergedAfter() throws {
    let date = Calendar.current.date(from: DateComponents(year: 2022, month: 9, day: 12))!
    let prs = try self.git.mergedPRs(.after(date: date))
    XCTAssertEqual(prs, [5102, 5105, 5065, 5090, 5066, 5094])
  }

  func testMergedPrs_mergedAfter_theDayAfter() throws {
    let date = Date(timeIntervalSinceNow: 84600)
    let prs = try self.git.mergedPRs(.after(date: date))
    XCTAssertEqual(prs, [])
  }

  func testMergedPrs_mergedUntil_withValidCommitMessage() throws {
    let commitMessage = "docs: changelog.sh update [skip ci]"
    let prs = try self.git.mergedPRs(.until(commitMessage: commitMessage))
    XCTAssertEqual(prs, [5090, 5066, 5094])
  }

  func testMergedPrs_mergedUntil_withInvalidCommitMessage() throws {
    let commitMessage = "FAKE"
    let prs = try self.git.mergedPRs(.until(commitMessage: commitMessage))
    XCTAssertEqual(prs, [5094])
  }
}

extension GitTests {
  func createGitRepositoryForTesting() throws {
    _ = try self.shell.executeOrThrow("rm -rf .git")
    _ = try self.shell.executeOrThrow("git init")

    try [
      "initial commit",
      "second commit",
      "Merge branch 'develop' into chore/support-xcode-14",
      "docs: changelog.sh update [skip ci]",
      "Merge pull request #5102 from BendingSpoons/feat/enable-new-primary-toolbar-ux",
      "Merge pull request #5105 from BendingSpoons/feat/growth/SP-27475-live-chat/9-onboarding-tests",
      "Merge pull request #5065 from BendingSpoons/feat/growth/SP-27475-live-chat/7-onboarding",
      "docs: changelog.sh update [skip ci]",
      "Merge pull request #5090 from BendingSpoons/feat/growth/SP-27475-live-chat/8-tests",
      "refactor: lint",
      "Merge branch 'develop' into fix/sca/SP-28188-num-animated-texts",
      "Merge pull request #5066 from BendingSpoons/feat/music-library-list-updates",
      "Merge pull request #5094 from BendingSpoons/feat/growth/SP-28060-unsubscribe-fake-door/8-localized-strings",
    ]
    .enumerated()
    .forEach { i, message in
      _ = try self.shell.executeOrThrow(
        "echo \(i) > file.txt; git add file.txt; git commit -m \"\(message)\""
      )
    }
  }
}
