// Copyright © Bending Spoons Group. All rights reserved.

import XCTest
@testable import FingerPullRequest

class MarkdownDocumentTests: XCTestCase {}

// MARK: - Checkbox Tests

extension MarkdownDocumentTests {
  func testCheckbox_isParsed_whenUnchecked() {
    let markdown = "- [ ] Feature Flag is REMOTE 🧪"
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == false)
  }

  func testCheckbox_isParsed_whenChecked() {
    let markdown = "- [x] Feature Flag is REMOTE 🧪"
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == true)
  }

  func testCheckbox_isParsed_whenTitleContainsSpecialCharacters() {
    let markdown = "- [x] Feature Flag is *REMOTE* 🧪"
    let checkbox = MarkdownElements.Checkbox("Feature Flag is *REMOTE* 🧪")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == true)
  }

  func testCheckbox_isParsed_whenCheckedWithCapitalizedX() {
    let markdown = "- [X] Feature Flag is REMOTE 🧪"
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == true)
  }

  func testCheckbox_descriptionIsIgnored_whenParsed() {
    let markdown = "- [x] Feature Flag is REMOTE 🧪 *the description*"
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪", description: "fake")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == true)
  }

  func testCheckbox_theRightCheckpoint_isParsed() {
    let markdown = """
    - [x] Feature Flag is AREMOTE 🧪 *the description*
    - [ ] Feature Flag is REMOTE 🧪 *the description*
    - [x] Feature Flag is BREMOTE 🧪 *the description*
    """
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪", description: "the description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == false)
  }

  func testCheckbox_theLastIsSelected_withMultipleMatches() {
    let markdown = """
    - [ ] Feature Flag is REMOTE 🧪 *the description*
    - [ ] Feature Flag is REMOTE 🧪 *the description*
    - [x] Feature Flag is REMOTE 🧪 *the description*
    """
    let checkbox = MarkdownElements.Checkbox("Feature Flag is REMOTE 🧪", description: "the description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: checkbox) == true)
  }
}

// MARK: - TextInput Tests

extension MarkdownDocumentTests {
  func testTextInput_returnsNil_whenEmpty() {
    let markdown = """
    ### The title
    > The description
    -
    """
    let textInput = MarkdownElements.TextInput("The title", description: "The description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssert(document.getValue(of: textInput) == "")
  }

  func testTextInput_returnsInputText_whenASingleLineIsInserted() {
    let markdown = """
    ### The title
    > The description
    - The text input
    """
    let textInput = MarkdownElements.TextInput("The title", description: "The description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(document.getValue(of: textInput), "The text input")
  }

  func testTextInput_returnsInputText_whenTitleHasSpecialCharacters() {
    let markdown = """
    ### The *title*
    > The description
    - The text input
    """
    let textInput = MarkdownElements.TextInput("The *title*", description: "The description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(document.getValue(of: textInput), "The text input")
  }

  func testTextInput_returnsInputText_whenMultilineIsInserted() {
    let markdown = """
    ### The title
    > The description
    - line 1
    line 2
    line 3
    """
    let textInput = MarkdownElements.TextInput("The title", description: "The description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(document.getValue(of: textInput), "line 1\nline 2\nline 3")
  }

  func testTextInput_ignoreDescription_whenMaching() {
    let markdown = """
    ### The title
    > The description
    - The text input
    """
    let textInput = MarkdownElements.TextInput("The title", description: "Fake")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(document.getValue(of: textInput), "The text input")
  }

  func testTextInput_returnsLastDeclaration_whenTitleClashes() {
    let markdown = """
    ### The title
    > The description
    - The text input 1

    ### The title
    > The description
    - The text input 2
    """
    let textInput = MarkdownElements.TextInput("The title", description: "The description")
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(document.getValue(of: textInput), "The text input 2")
  }

  func testTextInput_returnsInputText_whenUnusualCharactersArePresent() {
    let markdown = """
    ### *This* **a** : ~is~ @a $text \"containing\" [unusual] `markdown` 'text'.
    > The description
    - *This*: ~is~ **@a** $text \"containing\" [unusual] `markdown` 'text'.
    """
    let textInput = MarkdownElements.TextInput(
      "*This* **a** : ~is~ @a $text \"containing\" [unusual] `markdown` 'text'.",
      description: "The description"
    )
    let document = MarkdownDocument(parsing: markdown)
    XCTAssertEqual(
      document.getValue(of: textInput),
      "*This*: ~is~ **@a** $text “containing” [unusual] `markdown` ‘text’."
    )
  }
}
