// Copyright © Bending Spoons Group. All rights reserved.

import XCTest
@testable import FingerPullRequest

class UpdateChangelogTests: XCTestCase {
  func testUpdateChangelog_updatesTheChangelog_withMultiplePRs() {
    let existingChangelog =
      """
      ## unreleased

      - Entry 3

      ## Other

      - Entry 2
      - Entry 1
      """
    var newChangelog: String?
    let environment: Environment = .unimplemented(
      fileSystem: .unimplemented(
        read: { _ in existingChangelog.data(using: .utf8)! },
        write: { _, data in newChangelog = String(data: data, encoding: .utf8)! }
      ),
      git: .unimplemented(
        mergedPRs: { _ in [6, 5, 4] },
        currentBranch: { "branch" },
        add: { _ in },
        commit: { _ in },
        push: { _ in },
        pull: { _ in }
      ),
      github: .unimplemented(fetchPullRequest: { .mock(number: $0, title: "Title", body: "Body") })
    )
    let fingerPullRequest = FingerPullRequest(environment: environment)
    try! fingerPullRequest.updateChangelog {
      return "Entry \($0.number)"
    }

    XCTAssertEqual(
      newChangelog!,
      """
      ## unreleased

      - Entry 6
      - Entry 5
      - Entry 4
      - Entry 3

      ## Other

      - Entry 2
      - Entry 1
      """
    )
  }

  func testUpdateChangelog_updatesTheChangelog_anEmptyUnreleasedSection() {
    let existingChangelog =
      """
      ## unreleased


      ## Other

      - Entry 3
      - Entry 2
      - Entry 1
      """
    var newChangelog: String?
    let environment: Environment = .unimplemented(
      fileSystem: .unimplemented(
        read: { _ in existingChangelog.data(using: .utf8)! },
        write: { _, data in newChangelog = String(data: data, encoding: .utf8)! }
      ),
      git: .unimplemented(
        mergedPRs: { _ in [6, 5, 4] },
        currentBranch: { "branch" },
        add: { _ in },
        commit: { _ in },
        push: { _ in },
        pull: { _ in }
      ),
      github: .unimplemented(fetchPullRequest: { .mock(number: $0, title: "Title", body: "Body") })
    )
    let fingerPullRequest = FingerPullRequest(environment: environment)
    try! fingerPullRequest.updateChangelog {
      return "Entry \($0.number)"
    }

    XCTAssertEqual(
      newChangelog!,
      """
      ## unreleased

      - Entry 6
      - Entry 5
      - Entry 4

      ## Other

      - Entry 3
      - Entry 2
      - Entry 1
      """
    )
  }
}

// MARK: - PullRequest + Mock

extension PullRequest {
  fileprivate static func mock(number: Int, title: String, body: String) -> Self {
    let user = User(
      id: 0,
      login: "User",
      userType: .user
    )

    let repo = Repo(
      id: 1,
      name: "Repo",
      fullName: "Name",
      owner: user,
      isPrivate: true,
      description: nil,
      isFork: false,
      htmlURL: "https://www.bendingspoons.com"
    )

    let linkRelation = Link.Relation(stringLiteral: "https://www.bendingspoons.com")

    return .init(
      number: number,
      title: title,
      body: body,
      user: user,
      assignee: nil,
      assignees: nil,
      createdAt: Date(),
      updatedAt: Date(),
      closedAt: nil,
      mergedAt: nil,
      head: .init(label: "", ref: "", sha: "", user: user, repo: repo),
      base: .init(label: "", ref: "", sha: "", user: user, repo: repo),
      state: .merged,
      isLocked: false,
      isMerged: true,
      commitCount: nil,
      commentCount: nil,
      reviewCommentCount: nil,
      additions: nil,
      deletions: nil,
      changedFiles: nil,
      milestone: nil,
      htmlUrl: "https://www.bendingspoons.com",
      draft: false,
      links: Link(
        self: linkRelation,
        html: linkRelation,
        issue: linkRelation,
        comments: linkRelation,
        reviewComments: linkRelation,
        reviewComment: linkRelation,
        commits: linkRelation,
        statuses: linkRelation
      )
    )
  }
}
