// Copyright © Bending Spoons Group. All rights reserved.

import Foundation
import XCTest
@testable import FingerPullRequest

final class ShellTests: XCTestCase {
  // MockPipe doesn't work on Linux
  #if os(macOS)
    func testExecute_executesTheCommand() throws {
      for logsSensitiveInformation in [true, false] {
        for logFile in [URL(fileURLWithPath: "file.log"), nil] {
          var logCalls = 0
          var commandID = 0
          let command = "command arg1 arg2"
          let terminationStatus: Int32 = 0
          let output = "output message"
          let error = "error message"
          let workDir = URL(fileURLWithPath: ".")
          let outputPipe = MockPipe(mockContent: output)
          let errorPipe = MockPipe(mockContent: error)
          let redirectSuffix = logFile.map { " 2>&1 | tee -a \($0.path)" } ?? ""

          let shell: Shell = .bash(
            in: workDir,
            processBuilder: {
              MockProcess(
                mockTerminationStatus: terminationStatus,
                expectedArguments: [
                  "-eo",
                  "pipefail",
                  "-c",
                  "FALSE=0 ONE=1 SECRET=A_SECRET STRING=NOT_A_SECRET TRUE=1 \(command)\(redirectSuffix)",
                ],
                expectedCurrentDirectoryURL: workDir,
                expectedStandardOutput: outputPipe,
                expectedStandardError: errorPipe
              )
            },
            outputPipeBuilder: { outputPipe },
            errorPipeBuilder: { errorPipe },
            log: { id, content, label, _, level, isSensitive in
              logCalls += 1
              switch logCalls {
              case 1:
                commandID = id
                XCTAssertEqual(
                  content,
                  "Starting `FALSE=0 ONE=1 SECRET=****** STRING=NOT_A_SECRET TRUE=1 command arg1 arg2`"
                )
                XCTAssertNil(label)
                XCTAssertEqual(level, .error)
                XCTAssertFalse(isSensitive)
              case 2:
                XCTAssertEqual(id, commandID)
                XCTAssertEqual(content, "output message")
                XCTAssertEqual(label, "STDOUT")
                XCTAssertEqual(level, .info)
                XCTAssertEqual(isSensitive, logsSensitiveInformation)
              case 3:
                XCTAssertEqual(id, commandID)
                XCTAssertEqual(content, "error message")
                XCTAssertEqual(label, "STDERR")
                XCTAssertEqual(level, .error)
                XCTAssertEqual(isSensitive, logsSensitiveInformation)
              default:
                XCTFail("log should not be called more than 3 times")
              }
            }
          )
          let result = shell.execute(
            command,
            environment: [
              "SECRET": .init(value: "A_SECRET", isSensitive: true),
              "STRING": "NOT_A_SECRET",
              "TRUE": true,
              "FALSE": false,
              "ONE": 1,
            ],
            outputLogLevel: .info,
            errorLogLevel: .error,
            logFile: logFile,
            logsSensitiveInformation: logsSensitiveInformation
          )
          XCTAssertEqual(result, .init(status: terminationStatus, output: output, error: error))
          XCTAssertEqual(logCalls, 3)
        }
      }
    }

    func testExecute_whenMultipleCallsOnSameShell_executesOnDifferentProcesses() throws {
      let command = "command arg1 arg2"
      let terminationStatus: Int32 = 0
      let output = "output message"
      let error = "error message"
      let workDir = URL(fileURLWithPath: ".")
      let outputPipe = MockPipe(mockContent: output)
      let errorPipe = MockPipe(mockContent: error)

      let shell: Shell = .bash(
        in: workDir,
        processBuilder: {
          MockProcess(
            mockTerminationStatus: terminationStatus,
            expectedArguments: ["-eo", "pipefail", "-c", command],
            expectedCurrentDirectoryURL: workDir,
            expectedStandardOutput: outputPipe,
            expectedStandardError: errorPipe
          )
        },
        outputPipeBuilder: { outputPipe },
        errorPipeBuilder: { errorPipe },
        log: { _, _, _, _, _, _ in }
      )
      let firstResult = shell.execute(command)
      XCTAssertEqual(firstResult, .init(status: terminationStatus, output: output, error: error))
      let secondResult = shell.execute(command)
      XCTAssertEqual(secondResult, .init(status: terminationStatus, output: output, error: error))
    }

    func testExecuteOrThrow_whenResultIsSuccess_doesNotThrow() throws {
      let command = "command arg1 arg2"
      let terminationStatus: Int32 = 0
      let output = "output message"
      let error = "error message"
      let workDir = URL(fileURLWithPath: ".")
      let outputPipe = MockPipe(mockContent: output)
      let errorPipe = MockPipe(mockContent: error)

      let shell: Shell = .bash(
        in: workDir,
        processBuilder: {
          MockProcess(
            mockTerminationStatus: terminationStatus,
            expectedArguments: ["-eo", "pipefail", "-c", command],
            expectedCurrentDirectoryURL: workDir,
            expectedStandardOutput: outputPipe,
            expectedStandardError: errorPipe
          )
        },
        outputPipeBuilder: { outputPipe },
        errorPipeBuilder: { errorPipe },
        log: { _, _, _, _, _, _ in }
      )
      let result = try shell.executeOrThrow(command)
      XCTAssertEqual(result, .init(status: terminationStatus, output: output, error: error))
    }

    func testExecuteOrThrow_whenResultIsError_throws() throws {
      let command = "command arg1 arg2"
      let terminationStatus: Int32 = 1
      let output = "output message"
      let error = "error message"
      let workDir = URL(fileURLWithPath: ".")
      let outputPipe = MockPipe(mockContent: output)
      let errorPipe = MockPipe(mockContent: error)

      let shell: Shell = .bash(
        in: workDir,
        processBuilder: {
          MockProcess(
            mockTerminationStatus: terminationStatus,
            expectedArguments: ["-eo", "pipefail", "-c", command],
            expectedCurrentDirectoryURL: workDir,
            expectedStandardOutput: outputPipe,
            expectedStandardError: errorPipe
          )
        },
        outputPipeBuilder: { outputPipe },
        errorPipeBuilder: { errorPipe },
        log: { _, _, _, _, _, _ in }
      )

      XCTAssertThrowsError(
        try shell.executeOrThrow(command),
        expectedError: Shell.Error.commandFailed(command: command, errorMessage: "1: error message")
      )
    }
  #endif

  func testResultOutputTrimmed_trimsWhitespacesAndNewlines() throws {
    XCTAssertEqual(Shell.Result(status: 0, output: "output", error: "").outputTrimmed, "output")
    XCTAssertEqual(Shell.Result(status: 0, output: " output \n", error: "").outputTrimmed, "output")
  }

  func testResultErrorTrimmed_trimsWhitespacesAndNewlines() throws {
    XCTAssertEqual(Shell.Result(status: 1, output: "", error: "error").errorTrimmed, "error")
    XCTAssertEqual(Shell.Result(status: 1, output: "", error: " error \n").errorTrimmed, "error")
  }

  func testResultMessage_whenSuccess_printsStatusCode() throws {
    XCTAssertEqual(Shell.Result(status: 0, output: "", error: "error").resultMessage, "0")
  }

  func testResultMessage_whenFailure_printsStatusCodeAndErrorTrimmed() throws {
    XCTAssertEqual(Shell.Result(status: 1, output: "", error: " error \n").resultMessage, "1: error")
  }

  func testResultIsSuccess() throws {
    XCTAssertTrue(Shell.Result(status: 0, output: "asad", error: " error \n").isSuccess)
    XCTAssertFalse(Shell.Result(status: 1, output: "asad", error: " error \n").isSuccess)
    XCTAssertFalse(Shell.Result(status: -1, output: "asad", error: " error \n").isSuccess)
  }

  func testResultIsFailed() throws {
    XCTAssertFalse(Shell.Result(status: 0, output: "asad", error: " error \n").isFailed)
    XCTAssertTrue(Shell.Result(status: 1, output: "asad", error: " error \n").isFailed)
    XCTAssertTrue(Shell.Result(status: -1, output: "asad", error: " error \n").isFailed)
  }
}

#if os(macOS)
  class MockProcess: Process {
    var mockTerminationStatus: Int32
    var expectedArguments: [String]
    var expectedCurrentDirectoryURL: URL
    var expectedStandardOutput: Pipe
    var expectedStandardError: Pipe

    init(
      mockTerminationStatus: Int32,
      expectedArguments: [String],
      expectedCurrentDirectoryURL: URL,
      expectedStandardOutput: Pipe,
      expectedStandardError: Pipe
    ) {
      self.mockTerminationStatus = mockTerminationStatus
      self.expectedArguments = expectedArguments
      self.expectedCurrentDirectoryURL = expectedCurrentDirectoryURL
      self.expectedStandardOutput = expectedStandardOutput
      self.expectedStandardError = expectedStandardError
      super.init()
    }

    private var launchCalls = 0
    private var mockLaunchPath: String?
    private var mockArguments: [String]? // swiftlint:disable:this discouraged_optional_collection
    private var mockCurrentDirectoryURL: URL?
    private var mockStandardOutput: Any?
    private var mockStandardError: Any?

    override var launchPath: String? {
      get { return self.mockLaunchPath }
      set { self.mockLaunchPath = newValue }
    }

    override var arguments: [String]? { // swiftlint:disable:this discouraged_optional_collection
      get { return self.mockArguments }
      set { self.mockArguments = newValue }
    }

    override var currentDirectoryURL: URL? {
      get { return self.mockCurrentDirectoryURL }
      set { self.mockCurrentDirectoryURL = newValue }
    }

    override var standardOutput: Any? {
      get { return self.mockStandardOutput }
      set { self.mockStandardOutput = newValue }
    }

    override var standardError: Any? {
      get { return self.mockStandardError }
      set { self.mockStandardError = newValue }
    }

    override var terminationStatus: Int32 {
      return self.mockTerminationStatus
    }

    override func launch() {
      self.launchCalls += 1
      XCTAssertEqual(self.launchCalls, 1)
      XCTAssertEqual(self.expectedArguments, self.arguments)
      XCTAssertEqual(self.expectedCurrentDirectoryURL, self.currentDirectoryURL)
      XCTAssertEqual(self.expectedStandardOutput, self.standardOutput as? Pipe)
      XCTAssertEqual(self.expectedStandardError, self.standardError as? Pipe)
    }

    override func waitUntilExit() {}
  }

  class MockPipe: Pipe {
    var mockContent: String

    init(mockContent: String) {
      self.mockContent = mockContent
      super.init()
    }

    /// Create a FileHandle that always has the whole `mockContent` as `availableData`
    override var fileHandleForReading: FileHandle {
      let mockFileHandle = MockFileHandle()
      mockFileHandle.mockContent = self.mockContent
      return mockFileHandle
    }
  }

  class MockFileHandle: FileHandle {
    var mockContent: String = ""

    override func readDataToEndOfFile() -> Data {
      return self.mockContent.data(using: .utf8)!
    }
  }
#endif
